module main;
import std.stdio, std.string, std.conv, std.traits, std.range, dshow;
import std.algorithm : min, findAmong;
import std.array : replace;

//first some parser combinators
struct PS { //parser state
	bool ok;
	char[] start;
	char[] rest;
}

PS start(char[] data) { return PS(true, data, data); }
PS fail(PS st) { return PS(false, st.start, st.rest); }
PS success(PS st, int nc) { return PS(true, st.start, st.rest[nc..$]); }

//following (corrected) grammar from http://msdn.microsoft.com/en-us/library/aa451348.aspx
PS b1(PS st)
{
	if (!st.ok) return st;
	if (st.rest.length >= 2) {
		wchar wc = (cast(wchar[])st.rest[0..2])[0];
		if (" \t\r\n\0"w.indexOf(wc) >= 0) 
			return success(st, 2);	
	}
	return fail(st);
}

PS b(PS st)
{
	auto r = b1(st);
	if (!r.ok) return r;
	while(r.ok) r = b1(r);
	return PS(true, st.start, r.rest);
}

PS num(PS st, ref int r)
{
	if (!st.ok) return st;
	wchar[] ws = cast(wchar[])st.rest[0.. $ - ($&1)];
	int k = 0; r = 0;
	foreach(c; ws) {
		if (c >= '0' && c <= '9') {
			r = r*10 + cast(ubyte)c - 48;
			k++;
		} else break;
	}
	if (k>0) return success(st, k*2);
	return fail(st);
}

PS p_between(PS st, wchar start, wchar end, ref wstring str)
{
	if (!st.ok) return st;
	wchar[] ws = cast(wchar[])st.rest[0.. $ - ($&1)];
	if (ws.length >= 2 && ws[0] == start) {
		auto i = ws[1..$].indexOf(end);
		if (i >= 0) {
			str = to!wstring(ws[1..i+1]);
			return success(st, 4+i*2);		
		}
	}
	return fail(st);
}

PS p_name (PS st, ref wstring str) { return p_between(st, '"', '"', str); }
PS p_clsid(PS st, ref wstring str) { return p_between(st, '{', '}', str); }

PS p_file(PS st, ref wstring str)
{
	if (!st.ok) return st;
	wchar[] ws = cast(wchar[])st.rest[0.. $ - ($&1)];
	foreach(word; ["SOURCE"w, "SINK"w]) {
		auto n = word.length;
		if (ws[0..n] == word) {
			auto r = PS(true, st.start, st.rest[n*2..$]).b().p_name(str).b();
			str = word ~ ": " ~ str;
			return r;
		}
	}
	return fail(st);
}

PS opt(alias f)(PS st)
{
	if (!st.ok) return st;
	auto r = f(st);
	if (r.ok) return r;
	return st;
}

struct Filter {
	int n;
	wstring name, clsid, file;
	char[] data;
}

PS p_filter(PS st, ref Filter flt)
{
	if (!st.ok) return st;
	int n, datalen;
	wstring name, cls, fname;
	auto r = st.b().num(n).b().p_name(name).b().p_clsid(cls).b();
	r = r.opt!((x) { return x.p_file(fname); }).num(datalen).b1();
	if (r.ok && r.rest.length >= datalen) {
		flt.n = n; flt.name = name; flt.clsid = cls; flt.file = fname;
		flt.data = r.rest[0..datalen];
		return success(r, datalen);
	}
	return fail(st);
}

PS p_many(alias f, T)(PS st, ref T[] arr)
{
	if (!st.ok) return st;
	arr = new T[0];
	T x;
	auto r = f(st, x);
	while(r.ok) {
		arr ~= x;
		r = f(r, x);
	}
	if (arr.length > 0) return PS(true, st.start, r.rest);
	return fail(st);
}

string findName(string[] uuids, wstring guid)
{
	string g = to!string(guid);
	foreach(line; uuids)
		if (line.toUpper().indexOf(g) >= 0) 
			return line;
	return g;
}

struct Connection {
	int n1, n2, sample_size, fixed_size, temp_compr;
	wstring pin1, pin2, majortype, subtype, format_type;
	char[] format;

	void print(wstring[int] filter_names, string[] uuids)
	{
		writeln(filter_names[n1], ".", pin1, " --> ", filter_names[n2], ".", pin2);
		writeln("fixed size: ", fixed_size, ", temporal: ", temp_compr, ", sample size: ", sample_size);
		writeln("major type: ", findName(uuids, majortype));
		writeln("sub type: ", findName(uuids, subtype));
		writeln("format type: ", findName(uuids, format_type), "\nformat size: ", format.length);
		auto flen = format.length;
		if (format_type == "05589F80-C356-11CE-BF01-00AA0055595A" && flen >= VIDEOINFOHEADER.sizeof) {
			VIDEOINFOHEADER* vih = cast(VIDEOINFOHEADER*)&format[0];
			vih.print();
		} else
		if (format_type == "F72A76A0-EB0A-11d0-ACE4-0000C0CC16BA" && flen >= VIDEOINFOHEADER2.sizeof) {
			VIDEOINFOHEADER2* vih = cast(VIDEOINFOHEADER2*)&format[0];
			vih.print();
		} else
		if (format_type == "05589F81-C356-11CE-BF01-00AA0055595A" && flen >= WAVEFORMATEX.sizeof) {
			WAVEFORMATEX* wf = cast(WAVEFORMATEX*)&format[0];
			printRecord(*wf);
		} else
		if (format_type == "05589F82-C356-11CE-BF01-00AA0055595A" && flen >= MPEG1VIDEOINFO.sizeof) {
			MPEG1VIDEOINFO* vi = cast(MPEG1VIDEOINFO*)&format[0];
			vi.print();
		} else
		if (format_type == "E06D80E3-DB46-11CF-B4D1-00805F6CBBEA" && flen >= MPEG2VIDEOINFO.sizeof) {
			MPEG2VIDEOINFO* vi = cast(MPEG2VIDEOINFO*)&format[0];
			vi.print();
		}
		writeln("");
	}
}

PS p_conn(PS st, ref Connection con)
{
	if (!st.ok) return st;
	int n1, n2, sampsize, fixedsize, tempcomp, length, fmtlen;
	wstring name1, name2, major, subtype, fmttype;
	auto r = st.b().num(n1).b().p_name(name1).b().num(n2).b().p_name(name2).b()
		.num(sampsize).b().p_clsid(major).b().p_clsid(subtype).b()
		.num(fixedsize).b().num(tempcomp).b().num(length).b().p_clsid(fmttype).b()
		.num(fmtlen).b1();
	if (r.ok && r.rest.length >= fmtlen) {
		con.n1 = n1; con.n2 = n2; con.pin1 = name1; con.pin2 = name2;
		con.majortype = major; con.subtype = subtype; con.sample_size = length;
		con.fixed_size = fixedsize; con.temp_compr = tempcomp; con.format_type = fmttype;
		con.format = r.rest[0..fmtlen];
		return r.success(fmtlen);
	}
	return fail(st);
}

int main(string[] argv)
{	
	if (argv.length < 2) {
		writeln("usage: grfdump file.grf");
		return 1;
	}	

	auto f = File(argv[1],"rb");
	auto sz = cast(uint)f.size;
	char[] buf = new char[sz];
	buf = f.rawRead(buf);
	f.close();
	
	string s1 = cast(string)"FILTERS"w;
	auto i = buf.indexOf(s1);
	auto data = buf[i + s1.length .. $];
	Filter[] fts;
	auto r = start(data).p_many!(p_filter, Filter)(fts).b();
	if (!r.ok) {
		writeln("filters parsing failed"); return 1;
	}
	wstring[int] filter_names;
	writeln("Filters:");
	foreach(flt; fts) {
		writeln(flt.n, " ", flt.name, " ", flt.clsid, " ", flt.file, " filter data: ", flt.data.length, " bytes.");
		filter_names[flt.n] = flt.name;
	}
	auto con = "CONNECTIONS"w;
	wchar[] ws = cast(wchar[])r.rest[0..$-($&1)];	
	if (ws[0..con.length] != con) {
		writeln("connections parsing failed"); return 1;
	}
	Connection[] cons;
	r = start(r.rest[con.length*2..$]).p_many!(p_conn, Connection)(cons);
	if (!r.ok) {
		writeln("connections parsing failed"); return 1;
	}	
	string[] uuids = uuids_text.splitLines();
	writeln("\nConnections:");
	foreach(conn; cons)
		conn.print(filter_names, uuids);
	return 0;
}
